/**
 * indexer
 * Copyright(c) 2011 Jesus A. Domingo <jesus.domingo@gmail.com>
 * MIT Licensed
 */

/**
 * Constructor
 */
function Indexer(options) {
  
  this._values = [];
  this._keymap = [];
  
  if (typeof options == 'object') {

    this.name = options.name;
    this._type = (options.type === 'numeric' ? 'numeric' : 'string');
    this._length = parseInt(options.length) || 0

  } else {
    
    this.name = options;
    this._type = 'string';
    this._length = 0;
  
  }
}

/**
 * Search the index for all keys having this value
 */
Indexer.prototype.search = function (value, limit, offset) {

  if (typeof limit != 'number') {
    limit = 0;
  }
  if (typeof offset != 'number') {
    offset = 0;
  }
  offset = (offset < 0 ? 0 : offset);
  limit = (limit < 0 ? 0 : limit);

  if (typeof value == 'object' && (false == value instanceof Array)) {
    return this._searchRange(value, limit, offset);
  }

  return this._searchValues(value, limit, offset);
};

/**
 * Enumerate all keys in order
 */
Indexer.prototype.all = function (limit, offset) {
  
  if (typeof limit != 'number') {
    limit = 0;
  }
  if (typeof offset != 'number') {
    offset = 0;
  }
  offset = (offset < 0 ? 0 : offset);
  limit = (limit < 0 ? 0 : limit);

  var results = this._siftKeys(this._keymap);

  return (limit > 0 ? results.slice(offset, offset + limit) : results);
};

/**
 * Add a value and its key to the index
 */
Indexer.prototype.add = function (value, key) {

  if (this._type == 'string' && this._length != 0) {
    value = value.substr(0, this._length);
  }

  var meta = this._getValuePosition(value);

  if (meta.found) {
    this._keymap[meta.position].push(key);
  } else {
    this._values.splice(meta.position, 0, value);
    this._keymap.splice(meta.position, 0, [key]);
  }
};

/**
 * This removes a key from the index
 */
Indexer.prototype.remove = function (key) {
  for (var i = 0, n = this._keymap.length, p = 0; i < n; ++i) {
    p = this._keymap[i].indexOf(key);
    if (p >= 0) {
      delete this._keymap[i].splice(p, 1);
    }
  }
};

/**
 * Flushes keys for an indexed value
 */
Indexer.prototype.flush = function (value) {

  if (this._type == 'string' && this._length != 0) {
    value = value.substr(0, this._length);
  }

  var p = this._values.indexOf(value);
  
  if (p >= 0) {
    delete this._values.splice(p, 1);
    delete this._keymap.splice(p, 1);
  }
};

/**
 * Utility function for sifting result sets
 * of duplicate keys
 * @param {Array} 2D array of keys that have duplicates
 * @return {Array} 1D array of unique keys
 * @api private
 */
Indexer.prototype._siftKeys = function (keymap) {

  var i = j = n = m = 0,
      sieve = {},
      order = [];

  keymap.forEach(function (keylist) {
    keylist.forEach(function (key) {
      if (!sieve[key]) {
        sieve[key] = 1;
        order.push(key);
      }
    });
  });

  return order;
};

/**
 * Search the values using a min-max range
 * @param {Object} query object must contain 1 of [min, max]
 * @param {Number} optional, specify how many results max to get
 * @param {Number} optional, offset the results by this much
 * @return {Array} array of results
 * @api private
 */
Indexer.prototype._searchRange = function (opts, limit, offset) {

  var minVal = opts.min || this._values[0],
      maxVal = opts.max || (this._values[this._values.length - 1]);

  if (minVal > maxVal) {
    return [];
  }

  minPos = this._getValuePosition(minVal).position;
  maxPos = this._getValuePosition(maxVal).position + 1;

  var results = this._siftKeys(this._keymap.slice(minPos, maxPos));
  
  return (limit > 0 ? results.slice(offset, offset + limit) : results);
};

/**
 * Search for a list of values in the index
 * @param {Array} value or array of values to search
 * @param {Number} optional, specify how many results max to get
 * @param {Number} optional, offset the results by this much
 * @return {Array} array of results
 * @api private
 */
Indexer.prototype._searchValues = function (value, limit, offset) {

  if (false == value instanceof Array) {
    value = [value];
  }

  value.sort();

  var self = this,
      results = [],
      trim = function (s) { return s; };

  if (self._type == 'string' && self._length != 0) {
    trim = function (s) { return s.substr(0, self._length); };
  }

  value.forEach(function (val) {
    var meta = self._getValuePosition(trim(val));
    if (meta.found) {
      results.push(self._keymap[meta.position]);
    }
  });
  
  results = self._siftKeys(results);

  return (limit > 0 ? results.slice(offset, offset + limit) : results);
};

/**
 * Wrapper for the two search functionalities
 */
Indexer.prototype._getValuePosition = function (value) {

  var values = this._values;

  var low = 0,
      high = values.length - 1,
      middle = 0,
      comparison = 0,
      offset = 0,
      found = false;

  if (this._type == 'string') {
    value = value.toLowerCase();
  }

  while (low <= high) {

    middle = Math.floor((low + high) / 2);

    if (this._type == 'string') {
      comparison = values[middle].toLowerCase().localeCompare(value);
    } else {
      comparison = values[middle] - value;
    }

    if (comparison != 0) {
      if (comparison < 0) {
        low = middle + 1;
        offset = 1;
      } else {
        high = middle - 1;
        offset = 0;
      }
      continue;
    }

    found = true;
    offset = 0;
    
    break;
  }

  return { position: middle + offset, found: found };
};

/**
 * Create an index table
 */
exports.create = function (options) {
  return new Indexer(options);
};
